from rest_framework.viewsets import ModelViewSet
from rest_framework.response import Response

from rest_framework import serializers
from .methods import set_serializer, get_key_value_filter, select_fields
from .utils import mix_dicts_self_params, takeoff_list, remove_empty


class CountSerializer(serializers.Serializer):
    quantity = serializers.IntegerField()


class UrlFilterModelViewSet(ModelViewSet):
    # HIGHLY IMPORTANT NOTE: DON'T PERFORM "UPDATE" TO THESE, OR ALL YOUR REQUEST WILL APPLY IT!!!!! D:
    default_filter = {}  # example: {'field1': 'value1', 'field2': 'value2'}
    aliases = {}  # example: {'key_to_show1': 'key_to_search1', 'key_to_show2': 'key_to_search2'}
    serializer_base = {}  # example: {'model': Class, 'base': ClassSerializer}

    def list(self, request, *args, **kwargs):
        quantity = request.query_params.get('quantity', 'false').lower() == 'true'
        many = not quantity
        self.serializer_class = CountSerializer if quantity else set_serializer(self, request)
        filtered_list = self.qp_filter(self.queryset, request)
        serialize = self.serializer_class(filtered_list, many=many)
        return Response(serialize.data)

    def qp_filter(self, query_obj, request,
                  default_filter=None,
                  aliases=None,
                  selected_fields=None,
                  perform='filter',
                  priority_perform='self',
                  priority_filter='params',
                  priority_aliases='params'):
        """
        :param selected_fields: specify the fields to be filtered | Example: ['filter_me', 'and_me'] & ['-not_me']
        :param priority_perform: set priority to define method to perform | Values: 'self' & 'params'
        :param priority_aliases: set priority to join the aliases | Values: 'self' & 'params'
        :param priority_filter: set priority to join the default filters | Values: 'self' & 'params'
        :param perform: type of query | Values: 'filter' & 'get'
        :param aliases: switch keys between search and display
        :param default_filter: custom filter (receives a dict)
        :param query_obj: sort of "model.objects" (comes without "all()", "filter()" etc.)
        :param request: request with filter options at query_params
        :return: filtered query, or default if there's no query params
        """

        default_filter = mix_dicts_self_params(self=self.default_filter, params=default_filter,
                                               priority=priority_filter)

        if not len(request.query_params) and not default_filter:  # if there's nothing to filter, return all
            return query_obj.all()

        aliases = mix_dicts_self_params(self=self.aliases, params=aliases, priority=priority_aliases)

        query_params = takeoff_list(dict(request.query_params))  # make it clean and editable
        # For URL query params take priority
        if priority_filter == 'params':
            default_filter.update(query_params)
            query_params = default_filter
        elif priority_filter == 'self':
            query_params.update(default_filter)

        # # if "id" is single value, load "/<id>" with same parameters
        # # uncomment this if you don't have "ad-hoc methods" or if you know what you're doing
        # single_id = single_value(query_params['id'])
        # if single_id:
        #     params = url_encoder(query_params)
        #     return redirect(request.path + '/' + query_params['id'] + ('?' + params if params else ''))

        ns = {'result_query': None, 'query_set': query_obj}  # namespace for exec

        if 'fields' in query_params:
            del query_params['fields']
        if 'exclude' in query_params:
            del query_params['exclude']

        query_params = select_fields(query_params, selected_fields)

        quantity = True if query_params.pop('quantity', 'false').lower() == 'true' else None

        # Need to take it off query_params to have no effect at "len(query_params)", so I used "raw_perform"
        raw_perform = query_params.pop('perform', None)
        if priority_perform == 'params' and raw_perform:
            perform = raw_perform
        else:
            del raw_perform

        distinct = to_exclude = to_filter = order_by = None

        if aliases:
            # aliases example: {'key_to_show1': 'key_to_search1', 'key_to_show2': 'key_to_search2'}
            for key, value in aliases.items():
                if key in query_params:
                    query_params[value] = query_params.pop(key)

        order_by_opts = query_params.pop('order_by', None)
        distinct_opts = query_params.pop('distinct', None)

        if isinstance(distinct_opts, str):  # Allow empty params
            # Example: ?distinct=field
            distinct = ''
            distinct_opts = remove_empty(distinct_opts.split(','))
            for i, item in enumerate(distinct_opts):
                distinct += ('"%s"' + (', ' if i + 1 != len(distinct_opts) else '')) % item
            # ")" is out for empty param case
            # The default "order_by" is breaking the request (don't know why, so I moved it to the result of the query)
            distinct = '.order_by().distinct(' + distinct + ')'
            # Final: .distinct("field1") or .distinct()

        if order_by_opts:
            # Example: ?order_by=field1,field2
            order_by = ''
            order_by_opts = remove_empty(order_by_opts.split(','))
            for i, item in enumerate(order_by_opts):
                order_by += ('"%s"' + (', ' if i + 1 != len(order_by_opts) else ')')) % item

            if order_by:
                order_by = '.order_by(' + order_by
            # Final: .order_by("field1", "field2")

        # Note: gotta pop all query options out of query_params before actually mount the filter
        if len(query_params):  # fields to filter
            if priority_perform == 'params':
                perform = query_params.pop('perform', perform)
            to_filter, to_exclude = get_key_value_filter(query_params, ns)
            # Empty params case (e.g. "?field1=&field2=")
            if to_filter:
                to_filter = (('.%s(**{%s})' % (perform, to_filter)) +
                             (('.exclude(**{%s})' % to_exclude) if to_exclude and perform == 'filter' else ''))

        # Mount query options in the string
        query_opts = '{distinct}{filter}'.format(**{
            'distinct': (distinct if distinct else ''),
            'filter': (to_filter if to_filter else '')
        })

        # Mount the code in string to execute
        exec('result_query = query_set' +
             (query_opts if query_opts else '.all()') +
             ('.count()' if quantity else '') +
             (('; result_query = result_query%s;' % order_by) if order_by else ''),
             ns)

        if isinstance(ns['result_query'], int):
            ns['result_query'] = {'quantity': ns['result_query']}

        return ns['result_query'] if ns['result_query'] is not None else query_obj



import re
from copy import copy


def mix_dicts_self_params(self: dict, params: dict, priority='params'):
    # Mix the "self" and the "params", with default priority to "params"
    self = copy(self) if self else {}  # Returning a copy because the update was making it default in all requests
    params = params if params else {}
    if priority == 'self':
        params.update(self)
        return params
    elif priority == 'params':
        self.update(params)
        return self


def str_match_type(string: str):
    cs = starts = ends = starts_cs = ends_cs = match_type = False
    can_be_str = False

    try:
        int(string)  # Check if can be integer
    except ValueError:
        try:
            float(string)  # Check if can be float
        except ValueError:
            try:
                if string not in ['True', 'False']:
                    raise ValueError  # Check if can be boolean
            except ValueError:
                can_be_str = True
    if can_be_str:
        if string.startswith('i-'):
            cs = True
            string = remove_character(remove_character(string, 'first'), 'first')

        if string.startswith('~'):
            starts = True
            string = remove_character(string, 'first')

        elif string.startswith('i~'):
            starts_cs = True
            # Removes 2 first characters
            string = remove_character(remove_character(string, 'first'), 'first')

        if string.endswith('~'):
            ends = True
            string = remove_character(string, 'last')

        elif string.endswith('~i'):
            ends_cs = True
            # Removes 2 last characters
            string = remove_character(remove_character(string, 'last'), 'last')

        if cs:
            match_type = 'iexact'
        elif starts and ends:
            match_type = 'contains'
        elif starts_cs and ends:
            match_type = 'icontains'
        elif starts:
            match_type = 'startswith'
        elif starts_cs:
            match_type = 'istartswith'
        elif ends:
            match_type = 'endswith'
        elif ends_cs:
            match_type = 'iendswith'

    return match_type, string


def takeoff_list(dict: dict):
    # If use same "dict", it repeats items if there's a special character at 0 (don't have a clue)
    new_dict = {}
    for key, value in dict.items():
        new_dict.update({key: value[0]})
    return new_dict


def list_or_single_value(string: str, pattern: str):
    result = remove_empty(string.split(pattern))
    return result[0] if len(result) == 1 else result


def url_encoder(dict: dict):
    query_params = ''
    for i, (key, value) in enumerate(dict.items()):
        n = 1 if isinstance(value, str) else (2 if value else 0)
        if n and key != 'id':
            query_params += ('&' if i > 0 else '') + key + '=' + value
    return query_params


def remove_empty(some_list: list):
    return [x for x in some_list if x]


def remove_character(string: str, pos):
    """
    :param string: a string to be treated
    :param pos: the position of the character you want to remove | Values = "first" or "last"
    :return: the string without the character
    """
    if pos == 'first':
        pos = True
    elif pos == 'last':
        pos = False
    return ''.join([x for i, x in enumerate(string) if i != (0 if pos else len(string) - 1)])


def check_parse(values):
    t_int = 'int'
    t_float = 'float'
    t_date = 'date'
    is_list = isinstance(values, list)
    is_bool = isinstance(values, bool) if not is_list else False

    if not is_bool:
        if not is_list:
            values = [values]

        for i, value in enumerate(values):
            if value.startswith('p-'):
                value = value.replace('p-', '')
                if value.startswith(t_int + '-'):
                    values[i] = int(value.replace(t_int + '-', ''))
                elif value.startswith(t_float + '-'):
                    values[i] = float(value.replace(t_float + '-', ''))
                elif value.startswith(t_date + '-'):
                    # YYYY-mm-ddTHH:MM(+ anything) | Example: 2016-02-13T17:23:46.030000Z and 2016-02-13T17:23
                    values[i] = re.match('[0-9]{4}(-[0-9]{2}){2} [0-9]{2}:[0-9]{2}',
                                         value.replace(t_date + '-', '').replace('T', ' ')).group(0)

    return values if is_list or is_bool else values[0]



from copy import copy

from rest_framework.serializers import HyperlinkedModelSerializer

from .utils import remove_empty, str_match_type, list_or_single_value, check_parse, remove_character


def get_key_value_filter(key_value, ns):
    # Receives {'field1': 'value1|value2', 'field2': 'value3,value4', 'field3': 'value5', 'field4': 'number'}
    # Returns this string:  "field1__range": ["value1", "value2"],
    #                       "field2__in": ["value3", "value4"],
    #                       "field3__icontains": "value5",
    #                       "field4": "number",
    #                       "field5": value_X  # A variable created in "ns" with the value
    to_filter = to_exclude = ''
    final_i = len(key_value)
    for i, (key, value) in enumerate(key_value.items()):

        if '.' in key:
            key = key.replace('.', '__')

        range_query = False  # Also filters boolean indirectly

        is_str = isinstance(value, str)
        is_bool = False if is_str else isinstance(value, bool)

        if isinstance(value, str):
            # Range query case: "|", or Pipe
            # Multi value case: ",", or Comma
            # "value1|value2" or "value1,value2" turn into ["value1","value2"];
            # "value1," or "value1|" into ["1"]
            value = list_or_single_value(value, '|' if '|' in value else ',')
            if isinstance(value, list):
                is_str = is_bool = False
            range_query = False if is_bool or is_str or len(value) != 2 else True

        value = check_parse(value)

        # Or use the types of match queries ("contains", "startswith" etc.) or "range query"
        match_type, value = str_match_type(value) if not range_query else [False, value]
        # Set the name of the variable for exec's namespace
        ns_value = 'value_%s' % i
        # Handle single and multiple values mounting single query or "__in"/"__range" with a list
        key_value_string = (('"' + key +
                             ('__' + match_type if match_type else
                              ('__range' if range_query else
                               ('__in' if isinstance(value, list) else ''))) +
                             '": ' + ns_value +
                             (', ' if i + 1 != final_i else '')))

        if not is_bool and is_str and value.startswith('-'):
            value = remove_character(value, 'first')
            to_exclude += key_value_string
        else:
            to_filter += key_value_string
        # Assign the value to the variable in namespace
        ns.update({ns_value: value})
    return to_filter, to_exclude


def set_serializer(instance, request):
    if 'base' in instance.serializer_base or instance.serializer_class:
        fields, exclude, raw_fields = [], [], []
        if 'fields' in request.query_params:
            raw_fields = remove_empty(request.query_params['fields'].split(','))
        for raw_field in raw_fields:
            if raw_field.startswith('-'):
                exclude.append(raw_field.replace('-', ''))
            else:
                fields.append(raw_field)

        instance.serializer_class = serializer_factory(model=(None if 'model' not in instance.serializer_base
                                                              else instance.serializer_base['model']),
                                                       # "base" is priority here
                                                       base=(instance.serializer_class
                                                             if 'base' not in instance.serializer_base
                                                             else instance.serializer_base['base']),
                                                       fields=tuple(fields), exclude=tuple(exclude))
    # assign it to "self.serializer_class"
    return instance.serializer_class


def serializer_factory(model=None, base=HyperlinkedModelSerializer, fields=None, exclude=None, order_by=None):
    # Source for this f***ing solution: http://stackoverflow.com/a/27468982/4694834
    attrs = {}
    if model:
        attrs.update({'model': model})
    # "fields" take priority over "exclude"
    if fields:
        if hasattr(base.Meta, 'exclude'):
            del base.Meta.exclude
        attrs.update({'fields': fields})
    if exclude and not fields:
        if hasattr(base.Meta, 'fields'):
            fields = list(base.Meta.fields)
            for field in exclude:
                fields.remove(field)
            attrs.update({'fields': tuple(fields)})
        else:
            attrs.update({'exclude': exclude})

    parent = (object,)
    if hasattr(base, 'Meta'):
        parent = (base.Meta, object)
    Meta = type('Meta', parent, attrs)
    if model:
        class_name = model.__name__ + 'Serializer'
    else:
        class_name = 'CustomSerializer'
    return type(base)(class_name, (base,), {'Meta': Meta,})


def select_fields(query_params, selected_fields):
    if selected_fields:
        dont_filter, do_filter = [], []
        for selected_field in selected_fields:  # Mount both selected and not selected fields
            if selected_field.startswith('-'):
                dont_filter.append(selected_field.replace('-', ''))
            else:
                do_filter.append(selected_field)
        new_query_params = copy(query_params)
        if do_filter or dont_filter:
            condition = 'key not in do_filter' if do_filter else 'key in dont_filter'  # Mount the string to eval
            for key, value in new_query_params.items():
                if eval(condition):
                    del query_params[key]

    return query_params


def filter_fields(objects: list, query_params):
    raw_fields = remove_empty(query_params['fields'].split(',')) if 'fields' in query_params else []

    result, exclude, fields = [], [], []

    for field in raw_fields:
        if field.startswith('-'):
            exclude.append(field.replace('-', ''))
        else:
            fields.append(field)

    for obj in objects:
        filtered_fields = {} if fields else obj.copy()
        for key, value in obj.items():
            if key in fields:
                filtered_fields[key] = obj[key]
            elif key in exclude and not fields:
                del filtered_fields[key]
        result.append(filtered_fields)

    return result



from distutils.core import setup

version = '1.3.0'
app_name = 'django-rest-query-params-filter'
repo = 'https://github.com/jourdanrodrigues/' + app_name

setup(
    name=app_name,
    packages=['query_params_filter'],
    version=version,
    description='Allow your API consumers to make queries through query parameters.',
    author='Jourdan Rodrigues',
    author_email='thiagojourdan@gmail.com',
    url=repo,
    download_url=repo + '/tarball/v' + version,
    keywords=['url', 'filter', 'query', 'params', 'python', 'django', 'rest_framework'],
    classifiers=[],
)
